;========================================================================================
; The system that I use to test things on an Arduino accepts characters from a terminal.
; Different microapps are started by entering a designated character which calls an init
; subroutine.  They each have their own command (single character) table. You can enter
; numbers in hex or decimal and then type a character to act on that number.  A control D
; is used to exit a microapp back to the main line where one can start another one.
; Commands for this microapp are:
; /	set the T1 mode+prescaler
; g	graph one item ... do the min/max measurements but print out every A2D conversion
; l	iterate over all ladder levels, and all PWM duty cycle settings reporting ripple
; p	iterate over all PWM duty cycle settings reporting ripple
; s	set ladder and PWM to a given value and print the result of A2D and ripple
; v	verify ladder by doing 's' on 0% and 100% for each ladder step
;========================================================================================

;--------------;-------------------------------------------------------------------------
; Test out a design idea for EDN
;
; The Arduino Nano that I use has it's pins labeled with it's own designation this is to
; remind me where to connect and what that pin can do.
;
;	NANO_D9		PB1	OC1A		PCINT1	-> 8-bit PWM w/ low-pass w/ or w/o ladder
;	NANO_A0		PC0	ADC0		PCINT8	-> R-2R ladder, #1
;	NANO_A1		PC1	ADC1		PCINT9	-> R-2R ladder, #2
;	NANO_A2		PC2	ADC2		PCINT10	-> R-2R ladder, #4
;	NANO_A7			ADC7			<- analog input
;
.equ	LadderPort = PORTC
.equ	LadderDir = DDRC
.equ	LadderPin = PINC
.equ	LadderBit = PC0		; required to be low three pins of a port
.equ	LadderMask = 7 << LadderBit

.equ	IdeaPort = PORTB
.equ	IdeaDir = DDRB
.equ	IdeaPin = PINB
.equ	IdeaBit = PB1
.equ	IdeaMask = 1 << IdeaBit

.equ	IdeaADC = 7

; GPIOR0 is a bit addressable register so we use it for flags
.equ	GraphIt = 0

;----------------------------------------------------------------------------------------
; Our commands
;
IdeaCommands:
	Entry	EOT,	0,	IdeaExit	; SUB  ^D
	Entry	'/', 	0, 	SetPrescaleT1	; / (slash, solidus, divide)
	Entry	'g', 	0, 	GraphOneItem	; g
	Entry	'l', 	0, 	ScanLadder	; l
	Entry	'p', 	0, 	ScanPWM		; p
	Entry	's', 	0, 	ShowOneItem	; s
	Entry	'v', 	0, 	VerifyLadder	; v
	Entry	LINK,	LINK,	BasicCommands	; -> the standard commands

;---------------------------------------------------------------------------------------
; Stop timer 1 and exit the microapp.
;
IdeaExit:
	rcall	DisableT1
	rjmp	ExitApp

;---------------------------------------------------------------------------------------
; Initialize timer 1 and A2D converter to test PWM + R-2R design idea
;
IdeaInit:
	rcall	DisableT1

	rcall	Print
	.db	CR,LF,"Design Idea",CR,LF,0

	; Set up our command table
	ldi	r16,high(IdeaCommands<<1)
	sts	CmdH,r16
	ldi	r16,low(IdeaCommands<<1)
	sts	CmdL,r16

	;-----
	; All pins we use are set for output, and start out low
	;-----
	in	r16,LadderDir
	ori	r16,LadderMask
	out	LadderDir,r16

	in	r16,LadderPort
	andi	r16,~LadderMask
	out	LadderPort,r16

	sbi	IdeaDir,IdeaBit
	cbi	IdeaPort,IdeaBit

	;-----
	; Set up Timer1 for an 8-bit PWM
	;-----

.set	wgm = 5 ; Fast PWM 0 ... 0xFF		wgn = 1 ; Phase Correct 0 ... 0xFF adds additional ÷2
.set	clksel = 1 ; ÷1

	; Clear interrupt flags (by writing a one) and disable them all
	ldi	r16,(1<<ICF1) | (1<<OCF1B) | (1<<OCF1A) | (1<<TOV1)
	out	TIFR1,r16		
	ldi	r16, (0<<ICIE1) | (0<<OCIE1B) | (0<<OCIE1A) | (0<<TOIE1)
	sts	TIMSK1,r16

	; Start out at 0%
	ldi	r16,0
	sts	OCR1AH,r16
	sts	OCR1AL,r16
	sts	OCR1BH,r16
	sts	OCR1BL,r16

	; OCR1A clears on match and OCR1B normal pin, low two bits of wgm
	ldi	r16, (2<<COM1A0) | (0<<COM1B0) | ( (wgm&3)<<WGM10 )
	sts	TCCR1A,r16

	; rest of wgm and clock source
	ldi	r16, ( ((wgm>>2)&3)<<WGM12 ) | (clksel<<CS10)
	sts	TCCR1B,r16 	; Timer1 is now running

	;-----
	; Set up the analog-to-digital converter to use just the one analog input
	;-----
.set	prescale = 7 ; ÷128
.set	a2dclock = clock / (1<<prescale) ; must be between 50kHz and 200kHz

	; Power up the ADC
	lds	r16,PRR
	andi	r16,~(1<<PRADC)	; 0 powers up, 1 saves power
	sts	PRR,r16

	; we are only using one analog input so disable digital input for just that one pin
	; (unless it is one of the two exclusively analog pins)
	ldi	r16,(1<<IdeaADC)&((1<<ADC5D)|(1<<ADC4D)|(1<<ADC3D)|(1<<ADC2D)|(1<<ADC1D)|(1<<ADC0D))
	sts	DIDR0,r16

	; A2D not vcomp uses mux, when enabled start a conversion on timer1b match
	ldi	r16,(0<<ACME) | (5<<ADTS0)
	sts	ADCSRB,r16

	; VRef = Vcc, right aligned, our one and only analog input
	ldi	r16,(1<<REFS0) | (0<<ADLAR) | (IdeaADC<<MUX0)
	sts	ADMUX,r16

	; "A normal conversion takes 13 ADC clock cycles. The first conversion after the ADC is
	; switched on (ADEN in ADCSRA is set) takes 25 ADC clock cycles in order to initialize
	; the analog circuitry."  So lets get that out of the way right now.

	; enabled, start now, not auto trigger, interrupt cleared and disabled, 
	ldi	r16,(1<<ADEN) | (0<<ADSC) | (0<<ADATE) | (1<<ADIF) | (0<<ADIE) | (prescale<<ADPS0)
	sts	ADCSRA,r16
	ori	r16,(1<<ADSC)
	sts	ADCSRA,r16	; start the first conversion (just in case it hasn't yet been done)

InitWait:
	lds	r16,ADCSRA
	sbrc	r16,ADSC
	rjmp	InitWait

	ret

;---------------------------------------------------------------------------------------
; Set the timer 1's PWM frequency
;	A sets both the mode and the clock source
;
;					Source	Fast	Phase Correct
;	0	   ÷1	62.5kHz	 	1	x
;	1	   ÷2	31.25kHz	1		x
;	2	   ÷8	 7.8kHz	 	2	x
;	3	  ÷16	 3.9kHz		2		x
;	4	  ÷64	  976Hz	 	3	x
;	5	 ÷128	  488Hz		3		x
;	6	 ÷256	  244Hz	 	4	x
;	7	 ÷512	  122Hz		4		x
;	8	÷1024	   61Hz	 	5	x
;	9	÷2048	   30.5Hz	5		x
;
; Phase correct mode adds an additional divide by two.  The differenece between Fast & Phase
; Correct 8-bit modes is only WGM12 (=0 PC, =1 Fast) in TCCR1B
;
SetPrescaleT1:
	lds	r17,AHi
	tst	r17
	brne	BadPrescale

	lds	r17,ALow
	cpi	r17,10
	brsh	BadPrescale

	; Turn off all the bits we are concerned with at the moment
	lds	r16,TCCR1B
	andi	r16,~( (1<<WGM12) | (1<<CS12) | (1<<CS11) | (1<<CS10) )

	; Divide setting by two and If we were given an even number, turn on fast mode
	lsr	r17
	brcs	OddNumber
	ori	r16,(1<<WGM12)
OddNumber:

	; Select the clock source and fast/phase-correct mode
	; ( 0=off, 1..5 are internal prescaller sources)
	inc	r17	; 0..4 -> 1..5
	or	r16,r17
	sts	TCCR1B,r16

	; Show the result to confirm
	rcall	SendHex

	ret

BadPrescale:
	rcall	PrintRet
	.db	"Only 0-9 are allowed",CR,LF,0	

;---------------------------------------------------------------------------------------
; Set ladder and PWM and print A2D conversion result and scan for ripple min/max
;	The low byte of A is the PWM, the bottom three bits of the high byte is the ladder
;	GPIOR0,GraphIt is a flag that (if set) tells us to print out all A2D conversions
;	so something can try and draw the ripple
;
; Details, Details -- on a ATmegaxxx8 0x00 is not 0% but rather a one clock blip.  It
; appears that 0xFF is really a constant one, but I'm not taking chances.  For those
; two special values we switch the pin to normal output and clear or set it's value.
; Of course now when we go from 0x00 to 0x01 we jump from off to two clock ticks...
; Can't even it out unless we want to change the limit to 0xFE and subtract one from
; OCR1A... after testing for the two special cases...sigh.
;
GraphOneItem:
	call	SendCRLF
	sbi	GPIOR0,GraphIt
	rjmp	DoOneItem

ShowOneItem:
	cbi	GPIOR0,GraphIt
DoOneItem:

	;----
	; Select the ladder step and PWM duty cycle
	;----

	; First the ladder from the high byte
	lds	r16,AHi
	andi	r16,LadderMask

	in	r17,LadderPort
	andi	r17,~LadderMask

	or	r17,r16
	out	LadderPort,r17

	rcall	HexChar

	; then the PWM value from the low byte
	ldi	r16,0
	sts	OCR1AH,r16
	lds	r16,ALow
	sts	OCR1AL,r16

	; Clear the COM1A bits (= normal output)
	lds	r17,TCCR1A
	andi	r17, ~(3<<COM1A0)

	; Check for the special cases
	tst	r16
	breq	ZeroPercent
	cpi	r16,0xFF
	breq	OneHundredPercent

	; Not a special case, go to proper mode (fast clear on match, phase correct clear on up cnt)
	ori	r17,(2<<COM1A0)
	rjmp	SetPinMode

ZeroPercent:
	cbi	IdeaPort,IdeaBit
	rjmp	SetPinMode

OneHundredPercent:
	sbi	IdeaPort,IdeaBit
;	rjmp	SetPinMode -- fall through

SetPinMode:
	sts	TCCR1A,r17

	; Show the PWM setting
	rcall	SendHex
	call	SendSpace

	;-----
	; Do an A2D conversion and report the result
	;-----
	Pauseus	5000 ; µs  -- depends on the cap -- 0.01µF -> 500µs, 0.1µF -> 5ms

	; not auto trigger, start a conversion now
	lds	r16,ADCSRA
	andi	r16,~(1<<ADATE)
	ori	r16, (1<<ADSC)
	sts	ADCSRA,r16	; conversion is running (or soon will be)

	; wait for a conversion completion
A2DWait:
	lds	r16,ADCSRA
	sbrc	r16,ADSC
	rjmp	A2DWait

	; Grab the result and report it to the console
	lds	WL,ADCL
	lds	WH,ADCH
	rcall	SendHex16

;---------------------------------------------------------------------------------------
; Timer1b can be used to schedule an analog-to-digital conversion and since it is also
; timer1 that is used for the PWM which is the source of the ripple we should be able to
; measure (or even draw) the ripple by sweeping over repeated instances of the PWM cycle.
;
; FYI: I've nicknamed this technique "Slow-scilloscope™"
;
; At 16MHz system clock we need to set the analog prescaller to ÷128 which results in an
; analog clock of 125kHz. "The actual sample-and-hold takes place 1.5 ADC clock cycles
; after the start of a normal conversion..." "A normal conversion takes 13 ADC clock
; cycles." "When Auto Triggering is used, the prescaler is reset when the trigger event
; occurs. This assures a fixed delay from the trigger event to the start of conversion.
; In this mode, the sample-and-hold takes place two ADC clock cycles after the rising
; edge on the trigger source signal. Three additional CPU clock cycles are used for 
; synchronization logic."
;
;	ADC	ticks	seconds
;	 1.5	  192	 12µs
;	 2	  256	 16µs	= one full speed PWM cycle
;	13	1,664	104µs
;	13.5	1,728	108µs
;
; The max should be at the duty cycle setting (falling edge) and the min at the overflow
; (rising edge 0xFF) but minus the delays.  Since we scan the entire cycle all of that
; doesn't really matter, we'll find the min and max wherever they are.  In this specific
; case the sample-and-hold occurs exactly one PWM cycle after the trigger, ie  we should't
; see any offset between duty cycle setting and MaxAt and MinAt should be FF or 0. Phase
; correct changes that though.
;
; Note: A2D converter has an input bandwidth of around 38.5kHz.  This is really what limits
; our ability to encode repetitive wave forms, not the analog clock.
;
; Basically it'll take about 0.03s per duty cycle and 7s for all duty cycles and
; around 1 minute for a full ladder. (at full PWM speed)
;
; TODO: Maybe have a flag that says print out each measurement to be able to graph the ripple?
;
.def	MinH = r15	; the smallest A2D value found during a scan of the entire cycle ...
.def	MinL = r14
.def	MaxH = r13	; ... and the largest
.def	MaxL = r12
.def	MinAt = r11	; At which time index did we find the minimum ...
.def	MaxAt = r10	; ... and maximum A2D values.
.def	count = r22

	; Initialize min and max values
	;   MinAt := 0
	;   MaxAt := 0
	;   Max := 0
	;   Min := 0xFFFF
	ldi	r16,0
	mov	MinAt,r16
	mov	MaxAt,r16
	mov	MaxH,r16
	mov	MaxL,r16

	ldi	r16,0xFF
	mov	MinH,r16
	mov	MinL,r16

	; enable auto trigger
	lds	r16,ADCSRA
	ori	r16,(1<<ADATE)
	sts	ADCSRA,r16

	; loop over all 256 locations in a PWM cycle
	clr	count
MeasureLoop:

	; Check to see if any key was pressed and if so abort the loop
	call	GetRxCount
	breq	DontExit
	rjmp	MeasureExit

DontExit:

	; set the location where we want to start a conversion (which samples 2 analog clocks later)
	ldi	r16,0
	sts	OCR1BH,r16
	sts	OCR1BL,count

	; Clear the interrupt flags, by setting them to one.  The A2D conversion is triggered
	; by the rising edge of the OCF1B match interrupt flag so it must be cleared too.
	lds	r16,ADCSRA
	ori	r16,(1<<ADIF)
	sts	ADCSRA,r16

	ldi	r16,(1<<ICF1) | (1<<OCF1B) | (1<<OCF1A) | (1<<TOV1)
	out	TIFR1,r16		

	; wait for a conversion completion which is signaled with the interrupt flag
MeasureWait:
	lds	r16,ADCSRA
	sbrs	r16,ADIF
	rjmp	MeasureWait

	; Grab the result
	lds	WL,ADCL
	lds	WH,ADCH

	;-----
	; Show the result if requested
	;-----
	sbis	GPIOR0,GraphIt
	rjmp	SkipGraph

	call	SendSpace
	call	SendHex16	
SkipGraph:

	;-----
	; Compare with Min, save it and it's index if it is smaller
	;-----
	cp	MinH,WH
	brlo	NotSmaller	; branch if MinH < WH
	brne	IsSmaller	; branch if MinH > WH ie. WH < MinH
				; continue if MinH = WH

	cp	WL,MinL
	brsh	NotSmaller	; branch if WL >= MinL
				; continue if WL < MinL

IsSmaller:
	movw	MinH:MinL,WH:WL
	mov	MinAt,count

NotSmaller:

	;-----
	; Compare with Max, save it and it's index if it is larger
	;-----
	cp	WH,MaxH
	brlo	NotLarger	; branch if WH < MaxH
	brne	IsLarger	; branch if WH > MaxH
				; continue if WH = MaxH

	cp	MaxL,WL
	brsh	NotLarger	; branch if MaxL >= WL
				; continue if MaxL < WL ie. WL > MaxL

IsLarger:
	movw	MaxH:MaxL,WH:WL
	mov	MaxAt,count

NotLarger:

	;-----
	; And on to the next index
	;-----
	inc	count
	brne	MeasureLoop

	;-----
	; Report the results to the console
	;-----
	call	SendSpace
	movw	WH:WL,MinH:MinL
	call	SendHex16

	call	SendSpace
	mov	r16,MinAt
	call	SendHex

	call	SendSpace
	movw	WH:WL,MaxH:MaxL
	call	SendHex16

	call	SendSpace
	mov	r16,MaxAt
	call	SendHex

	call	SendCRLF

MeasureExit:
	ret

.undef	MinH
.undef	MinL
.undef	MaxH
.undef	MaxL
.undef	MinAt
.undef	MaxAt
.undef	count

;---------------------------------------------------------------------------------------
; Scan all duty cycle settings of timer1a discovering the min/max voltages for each
;
ScanWholePWM:
	call	SendCRLF

	ldi	r16,0
ScanLoop:
	sts	ALow,r16

	; Check to see if any key was pressed and if so abort the loop
	call	GetRxCount
	brne	ScanExit

	; do the necessary measurements for this dutycycle
	rcall	ShowOneItem	

	; on to the next dutycycle value
	lds	r16,ALow
	inc	r16
	brne	ScanLoop

ScanExit:
	ret

;---------------------------------------------------------------------------------------
; Scan all duty cycle settings for all ladder settings
;
ScanLadder:
	; Start at ladder setting 0
	ldi	r16,0
LadderLoop:
	sts	AHi,r16

	; Check to see if any key was pressed and if so abort the loop
	call	GetRxCount
	brne	LadderExit

	rcall	ScanWholePWM

	; On to the next ladder setting, if any
	lds	r16,AHi
	inc	r16
	cpi	r16,8
	brlo	LadderLoop

LadderExit:
	call	SendCRLF
	ret

;---------------------------------------------------------------------------------------
; Scan just once through the PWM, with no ladder in place
;
ScanPWM:
	ldi	r16,0
	sts	AHi,r16

	rcall	ScanWholePWM
	ret

;---------------------------------------------------------------------------------------
; Verify ladder by getting the voltages and ripple for 0% and 100% PWM for each ladder level
;
VerifyLadder:
	call	SendCRLF

	; Start at ladder setting 0
	ldi	r16,0
VerifyLoop:
	sts	AHi,r16

	; Check to see if any key was pressed and if so abort the loop
	call	GetRxCount
	brne	VerifyExit

	; set the PWM to 0%
	ldi	r16,0
	sts	ALow,r16

	rcall	ShowOneItem

	; set the PWM to 100%
	ldi	r16,0xFF
	sts	ALow,r16

	rcall	ShowOneItem

	; On to the next ladder setting, if any
	lds	r16,AHi
	inc	r16
	cpi	r16,8
	brlo	VerifyLoop

VerifyExit:
	call	SendCRLF
	ret

.exit

